/* gmp_pm1.c
 *
 * Implementation of Pollard's p-1 algorithm using the GNU MP library.
 */

#include "gmp_pm1.h"

/* Performs an iteration of pollards p - 1 algorithm.
 * 
 * d	This is where any divisor found is stored.
 * n	The number to be factored.
 * m	lcm(1,2,...,b)
 * b	The smoothness bound to be used.
 * a	The a that should be used to generate numbers.
 *  
 * Returns 0 if no factor could be found, 1 otherwise (in which case d 
 * contains a non-trivial factor).
 */
static int gmp_factor_pm1_iteration(mpz_t d, mpz_t n, mpz_t m, unsigned long b, 
unsigned int a) {
	mpz_t x, mpa;
	
	mpz_init(x);
	mpz_init_set_ui(mpa, a);
	mpz_powm(x, mpa, m, n);
	mpz_sub_ui(x, x, 1);
	mpz_gcd(d, x, n);
	
	if(mpz_cmp_ui(d, 1) > 0 && mpz_cmp(d, n) < 0) {
		// We found a non-trivial factor.
		mpz_clear(x);
		mpz_clear(mpa);
		return 1;
	} else {
		mpz_clear(x);
		mpz_clear(mpa);
		return 0;
	}
}

/* Factor the number n into a (not necessarily prime) factor p. At most max_a
 * elements are tested. max_a = 0 disables this limit. b is the bound used.
 * Returns 0 on success, or 1 on failure (out of time).
 */
int gmp_factor_pm1(mpz_t p, mpz_t n, unsigned long max_a, unsigned long b) {
	unsigned int a, i;
	mpz_t m;
	
	/* Compute m = lcm(1,2,...,m), can be done faster by sieving. */
	mpz_init_set_ui(m, 1);
	for(i = 1; i <= b; i++) {
		mpz_lcm_ui(m, m, i);
	}
	
	/* Perform the specified amount of iterations. */
	for(a = 2; max_a == 0 || a < max_a+2; a++) {
		if(gmp_factor_pm1_iteration(p, n, m, b, a)) {
			mpz_clear(m);
			return 0;
		}
	}
	mpz_clear(m);
	return 1;
}

