/* gmp_qs.c
 *
 * An implementation of the Quadratic Sieve using the GNU MP library.
 */

#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <stdint.h>
#include <math.h>

#include "gmp_qs.h"

/* This seems best on most modern 32-bit x86 machines */
#define SYSTEM_GENERIC		0
/* Tuned for KTH's Kattis system */
#define SYSTEM_KATTIS		1
/* Tested on a Sempron64 2800+ */
#define SYSTEM_AMD64		2

#define USE_SYSTEM SYSTEM_GENERIC

/* SIEVE_LOG_STEP	should be close to the L1 cache size
 * SIEVE_LEN 		has to be a multiple of 2^SIEVE_LOG_STEP
 * USE_PR		non-zero if partial relations are to be used
 */
#if USE_SYSTEM == SYSTEM_KATTIS
#define USE_PR		0
#define SIEVE_LEN	0x8000
#define SIEVE_LOG_STEP	14
#elif USE_SYSTEM == SYSTEM_AMD64
#define USE_PR		1
#define SIEVE_LEN	0x20000
#define SIEVE_LOG_STEP	15
#else
#define USE_PR		1
#define SIEVE_LEN	0x20000
#define SIEVE_LOG_STEP	14
#endif

/* An element in a sparse matrix.
 */
struct element {
	/* The column that the element belongs to. */
	unsigned int column;
	/* The next element (column-wise). NULL if there is none. */
	struct element *next;
	/* The element's row. */
	struct row *row;
};

/* A row in a sparse matrix.
 */
struct row {
	/* The first and last element in the row. NULL if there is none. */
	struct element *first;
	struct element *last;
	/* The next and previous row in the matrix. NULL if there is none. */
	struct row *next;
	struct row *prev;
	/* The row index corresponding to the row. It's only assigned if the
	 * row belongs to a matrix. */
	unsigned int index;
};

/* Sparse matrix representation. Elements that are not stored are assumed to
 * be 0 and the others are assumed to be 1.
 */ 
struct sparse {
	/* The first and last row.NULL if there is none. */
	struct row *first_row;
	struct row *last_row;
	/* The number of rows in the matrix. */
	unsigned int row_count;
};

/* A binary tree of partial relations */
typedef struct s_prtree {
	struct s_prtree *left, *right;
	/* Remainder after dividing away the factor base */
	mpz_t r;
	/* Square root mod n of the factored number */
	mpz_t num;
	/* Number of element in factoring */
	int len;
	/* Factoring over the factor base, in sparse representation */
	unsigned int col[];
} prtree;

/* The polynomial struct pretty much represents the entire state of the
 * sieving process, including parameters for the current polynomial used
 * and the factor base. */
typedef struct {
	/* Size of factor base */
	int plen;
	/* Factor base */
	unsigned long *p;
	/* Approximate 2-logarithms of the factor base primes */
	int_log *log_p;
	/* p^1 (mod n) */
	mpz_t *p_inv;
	/* Square roots of n modulo p_i in the factor base */
	unsigned long *sqrt_n_mod_p;
	/* ceil(sqrt(n)) mod p_i */
	unsigned long *sqrt_n;
	/* Polynomial to use: (2*a*x + b)^2 */
	mpz_t a, b;
	/* (2d)^-1, where d^2 = a */
	mpz_t d2_inv;
	/* (2*a)^-1 (mod p_i) for all p_i in the factor base */
	unsigned long *a_inv_mod_p;
	/* b (mod p_i) */
	unsigned long *b_mod_p;
	/* root[2*i] and root[2*i+1] are roots of the polynomial mod p_i */
	unsigned long *root;
	/* Value around which to sieve */
	mpz_t x0;
} polynomial;

/* Solve the modular equatioon x^2 = n (mod p) using the Shanks-Tonelli
 * algorihm. x will be placed in q and 1 returned if the algorithm is
 * successful. Otherwise 0 is returned (currently in case n is not a quadratic
 * residue mod p). A check is done if p = 3 (mod 4), in which case the root is
 * calculated as n ^ ((p+1) / 4) (mod p).
 */
static int mpz_sqrt_mod(mpz_t q, mpz_t n, mpz_t p) {
    mpz_t w, n_inv, y;
    unsigned int i, s;

    if(mpz_legendre(n, p) != 1)         /* Not a quadratic residue?         */
        return 0;                       /* No, so return error              */
    if(mpz_tstbit(p, 1) == 1) {         /* p = 3 (mod 4) ?                  */
        mpz_set(q, p);
        mpz_add_ui(q, q, 1);
        mpz_fdiv_q_2exp(q, q, 2);
        mpz_powm(q, n, q, p);           /* q = n ^ ((p+1) / 4) (mod p)      */
        return 1;
    }
    mpz_set(q, p);
    mpz_sub_ui(q, q, 1);                /* q = p-1                          */
    s = 0;                              /* Factor out 2^s from q            */
    while(mpz_tstbit(q, s) == 0) s++;
    mpz_fdiv_q_2exp(q, q, s);           /* q = q / 2^s                      */
    mpz_init_set_ui(w, 2);              /* Search for a non-residue mod p   */
    while(mpz_legendre(w, p) != -1)     /* by picking the first w such that */
        mpz_add_ui(w, w, 1);            /* (w/p) is -1                      */
    mpz_powm(w, w, q, p);               /* w = w^q (mod p)                  */
    mpz_add_ui(q, q, 1);
    mpz_fdiv_q_2exp(q, q, 1);           /* q = (q+1) / 2                    */
    mpz_powm(q, n, q, p);               /* q = n^q (mod p)                  */
    mpz_init(n_inv);
    mpz_invert(n_inv, n, p);
    mpz_init(y);
    for(;;) {
        mpz_powm_ui(y, q, 2, p);        /* y = q^2 (mod p)                  */
        mpz_mul(y, y, n_inv);
        mpz_mod(y, y, p);               /* y = y * n^-1 (mod p)             */
        i = 0;
        while(mpz_cmp_ui(y, 1) != 0) {
            i++;
            mpz_powm_ui(y, y, 2, p);    /*  y = y ^ 2 (mod p)               */
        }
        if(i == 0) {                    /* q^2 * n^-1 = 1 (mod p), return   */
            mpz_clear(y);
            mpz_clear(w);
            mpz_clear(n_inv);
            return 1;
        }
        if(s-i == 1) {                  /* In case the exponent to w is 1,  */
            mpz_mul(q, q, w);           /* Don't bother exponentiating      */
        } else {
            mpz_powm_ui(y, w, 1 << (s-i-1), p);
            mpz_mul(q, q, y);
        }
        mpz_mod(q, q, p);               /* r = r * w^(2^(s-i-1)) (mod p)    */
    }
}

static void polynomial_destroy(polynomial *poly) {
	int i;
	for(i=0; i<poly->plen; i++) mpz_clear(poly->p_inv[i]);
	mpz_clear(poly->a);
	mpz_clear(poly->b);
	mpz_clear(poly->d2_inv);
	mpz_clear(poly->x0);
	free(poly->p);
	free(poly->log_p);
	free(poly->p_inv);
	free(poly->sqrt_n_mod_p);
	free(poly->sqrt_n);
	free(poly->a_inv_mod_p);
	free(poly->b_mod_p);
	free(poly->root);
	free(poly);
}

/* Compute roots of the polynomial modulo the primes in the factor base */
static void polynomial_roots(polynomial *poly, long x) {
	int i;
	mpz_t x_sum, root0, root1;
	unsigned long p, b_a_inv, nsqrt_a_inv;

	mpz_init(x_sum);
	mpz_init(root0);
	mpz_init(root1);
	if(x < 0) mpz_sub_ui(x_sum, poly->x0, -x);
	else mpz_add_ui(x_sum, poly->x0, x);
	mpz_neg(x_sum, x_sum);
	for(i=2; i<poly->plen; i++) {
		p = poly->p[i];
		/* roots are: (-b mod p +/- sqrt(n) mod p)*(2a^-1 mod p) */
		b_a_inv = ((uint64_t)poly->b_mod_p[i] *
			(uint64_t)poly->a_inv_mod_p[i]) % (uint64_t)p;
		nsqrt_a_inv = ((uint64_t)poly->sqrt_n_mod_p[i] *
			(uint64_t)poly->a_inv_mod_p[i]) % (uint64_t)p;
		mpz_add_ui(root0, x_sum, p - b_a_inv + nsqrt_a_inv);
		mpz_add_ui(root1, x_sum, 2*p - b_a_inv - nsqrt_a_inv);
		mpz_mod_ui(root0, root0, p);
		mpz_mod_ui(root1, root1, p);
		poly->root[i*2] = mpz_get_ui(root0);
		poly->root[i*2+1] = mpz_get_ui(root1);
	}
	mpz_clear(x_sum);
	mpz_clear(root0);
	mpz_clear(root1);
}

/* Update parameters associated with the 'a' coefficient of the polynomial.
 * The caller should set poly->a and then call this. */
static void polynomial_update_a(polynomial *poly, mpz_t n) {
	int i;
	mpz_t a2, x;

	mpz_init(a2);
	mpz_init(x);

	/* Build a list of (2a)^-1 mod p_i for the factor base */
	mpz_mul_2exp(a2, poly->a, 1);
	for(i=2; i<poly->plen; i++) {
		mpz_set_ui(x, poly->p[i]);
		mpz_invert(x, a2, x);
		poly->a_inv_mod_p[i] = mpz_get_ui(x);
	}

	mpz_clear(a2);
	mpz_clear(x);
}

/* Update parameters associated with the 'b' coefficient of the polynomial.
 * The caller should set poly->b and then call this. */
static void polynomial_update_b(polynomial *poly) {
	int i;
	mpz_t x;

	/* Build a list of b mod p_i for the factor base */
	mpz_init(x);
	for(i=2; i<poly->plen; i++) {
		mpz_mod_ui(x, poly->b, poly->p[i]);
		poly->b_mod_p[i] = mpz_get_ui(x);
	}
	mpz_clear(x);
}

/* Update the polynomial, by finding new coefficients and updating relevant
 * tables. d should be initialized to approx. (n/(2m^2))^(1/4), and will be
 * updated. */
static void polynomial_update(polynomial *poly, mpz_t d, mpz_t n) {
	mpz_t r1, df, inv_df, fr1, fr1p, t, b, sqrt_n;
	int mod4;

	mpz_init(r1);
	mpz_init(df);
	mpz_init(inv_df);
	mpz_init(fr1);
	mpz_init(fr1p);
	mpz_init(t);
	mpz_init(b);
	mpz_init(sqrt_n);

	/* Generate a prime d = 3 (mod 4) where n is a quadratic residue */
	do {
		mpz_add_ui(d, d, 2);
		mpz_nextprime(d, d);
		mod4 = mpz_tstbit(d, 0) + 2*mpz_tstbit(d, 1);
	} while(mod4 != 3 || mpz_legendre(n, d) != 1);

	/* a = d^2 */
	mpz_mul(poly->a, d, d);
	polynomial_update_a(poly, n);

	/* Hensel's Lemma is used to find a 'b' coefficient such that
	 * b^2 = n mod a, where a = d^2 and d is prime
	 * http://en.wikipedia.org/wiki/Hensel's_lemma */
	mpz_sqrt_mod(r1, n, d);
	mpz_mul_2exp(df, r1, 1);
	mpz_invert(inv_df, df, d);
	mpz_neg(inv_df, inv_df);
	mpz_mul(fr1, r1, r1);
	mpz_sub(fr1, fr1, n);
	mpz_mod(fr1, fr1, poly->a);
	mpz_fdiv_q(fr1p, fr1, d);
	mpz_mul(t, fr1p, inv_df);
	mpz_mul(b, t, d);
	mpz_add(b, b, r1);
	mpz_mod(b, b, poly->a);
	if(mpz_even_p(b)) mpz_sub(b, poly->a, b);

	mpz_set(poly->b, b);

	/* Compute x0, the center of our sieving interval.
	 * (2*a*x0 + b)^2 ~ n, so x0 = (sqrt(n)-b)/2a */
	mpz_sqrt(sqrt_n, n);
	mpz_sub(poly->x0, sqrt_n, poly->b);
	mpz_fdiv_q(poly->x0, poly->x0, poly->a);
	mpz_fdiv_q_2exp(poly->x0, poly->x0, 1);
	mpz_add_ui(poly->x0, poly->x0, 1);

	polynomial_update_b(poly);

	/* Compute (2*d)^-1 mod n */
	mpz_mul_2exp(t, d, 1);
	mpz_invert(t, t, n);
	mpz_set(poly->d2_inv, t);

	mpz_clear(r1);
	mpz_clear(df);
	mpz_clear(inv_df);
	mpz_clear(fr1);
	mpz_clear(fr1p);
	mpz_clear(t);
	mpz_clear(b);
	mpz_clear(sqrt_n);
}

/* Initialize the state of a sieve, and return a polynomial struct.
 * n is the number to factor, and plen is the size of the factor base. */
static polynomial *polynomial_create(mpz_t n, int plen) {
	polynomial *poly;
	mpz_t x, modsqrt_n_x;
	int i;
	unsigned long logref, x_ui;
	int_log log;

	poly = malloc(sizeof(*poly));
	poly->p = malloc(plen*sizeof(*(poly->p)));
	poly->log_p = malloc(plen*sizeof(*(poly->log_p)));
	poly->p_inv = malloc(plen*sizeof(*(poly->p_inv)));
	poly->a_inv_mod_p = malloc(plen*sizeof(*(poly->a_inv_mod_p)));
	poly->b_mod_p = malloc(plen*sizeof(*(poly->b_mod_p)));
	poly->sqrt_n_mod_p = malloc(plen*sizeof(*(poly->sqrt_n_mod_p)));
	poly->sqrt_n = malloc(plen*sizeof(*(poly->sqrt_n)));
	poly->root = malloc(2*plen*sizeof(*(poly->root)));
	poly->plen = plen;

	mpz_init(poly->a);
	mpz_init(poly->b);
	mpz_init(poly->d2_inv);
	mpz_init(poly->x0);

	poly->p[0] = 0;		/* Actually, -1, but never used		*/
	poly->p[1] = 2;		/* Also not one of the "normal" primes	*/
	poly->log_p[0] = 1;
	poly->log_p[1] = 1;

	mpz_init(modsqrt_n_x);	/* Modular sqrt of n, mod x		*/
	mpz_init_set_ui(x, 3);	/* Current prime to be tested		*/
	i = 2;			/* Number of primes in list		*/
	log = 1;		/* log2(3) ~ 2, 1 will be added		*/
	logref = 3;		/* Limit where log2(x) increases	*/

	while(i < plen) {
		/* n quadratic residue mod x? */
		if(mpz_legendre(n, x) == 1) {
			x_ui = mpz_get_ui(x);
			/* Yes, store x in prime list */
			poly->p[i] = x_ui;
			mpz_init(poly->p_inv[i]);
			/* Compute x^-1 (mod n) */
			mpz_invert(poly->p_inv[i], x, n);
			while(x_ui >= logref) {
				/* x >= log2 increase treshold? */
				/* Increase log and double treshold */
				log++;
				logref *= 2;
			}
			/* Store log2(x) in list */
			poly->log_p[i] = log;
			/* Get x and the modular sqrt(n) in x as normal
			 * single-precision variables, to avoid unnecessary
			 * multiprecision calculations */
			mpz_sqrt_mod(modsqrt_n_x, n, x);
			poly->sqrt_n_mod_p[i] = mpz_get_ui(modsqrt_n_x);
			i++;
		}
		mpz_nextprime(x, x);
	}

	/* Set/compute the inverses of -1 and 2 */
	mpz_init(poly->p_inv[0]);
	mpz_sub_ui(poly->p_inv[0], n, 1);
	mpz_init_set_ui(poly->p_inv[1], 2);
	mpz_invert(poly->p_inv[1], poly->p_inv[1], n);

	mpz_clear(modsqrt_n_x);
	mpz_clear(x);

	return poly;
}

/* Add the element 1 to the specified column in the specified row. */
static void append_element(struct row *row, int column) {
	struct element *new;
	new = malloc(sizeof(*new));
	new->column = column;
	new->next = NULL;
	new->row = row;
	
	if(row->first == NULL) {
		row->first = new;
	}
	if(row->last != NULL) {
		row->last->next = new;
	}
	row->last = new;
}

/* Append the specified row to the specified matrix. */
static void append_row(struct sparse *matrix, struct row *row) {
	row->index = matrix->row_count++;
	if(matrix->first_row == NULL) {
		matrix->first_row = row;
	}
	if(matrix->last_row != NULL) {
		matrix->last_row->last->next = row->first;
		row->prev = matrix->last_row;
		matrix->last_row->next = row;
	}
	matrix->last_row = row;
}

/* Destroy a row and its elements. */
static void destroy_row(struct row *row) {
	if(row->prev != NULL) {
		/* There's a previous row that we need to relink. */
		row->prev->last->next = row->last->next;
		row->prev->next = row->next;
	}
	if(row->next != NULL) {
		/* There's a next row that we need to relink. */
		row->next->prev = row->prev;
	}
	
	/* Free the elements and row. */
	struct element *e = row->first;
	struct element *next = e;
	while(e != row->last) {
		next = e->next;
		free(e);
		e = next;
	}
	free(row->last);
	free(row);
}

#if USE_PR
/* Count the number of non-zero elements in the row. */
static int row_length(struct row *row) {
	if(row->first == NULL) return 0;
	int i = 1;
	struct element *e;
	for(e = row->first; e != row->last; e = e->next) {
		i++;
	}
	return i;
}
#endif

/* Destroy a sparse matrix, its rows and elements. */
static void destroy_sparse(struct sparse *sparse) {
	if(sparse->first_row != NULL) {
		struct element *e = sparse->first_row->first;
		struct element *next = e;
		while(e != NULL) {
			next = e->next;
			free(e);
			e = next;
		}
	
		struct row *r = sparse->first_row;
		struct row *next_row = r;
		while(r != NULL) {
			next_row = r->next;
			free(r);
			r = next_row;
		}
	}
	
	free(sparse);
}

/* Create a new independent row. */
static struct row *create_row() {
	struct row *row;
	row = malloc(sizeof(*row));
	row->first = row->last = NULL;
	row->next = row->prev = NULL;
	return row;
}

/* Create a new sparse matrix without any rows. */
static struct sparse *create_sparse() {
	struct sparse *matrix;
	matrix = malloc(sizeof(*matrix));
	matrix->row_count = 0;
	matrix->first_row = NULL;
	matrix->last_row = NULL;
	return matrix;
}

#if USE_PR
/* Crete a partial relation node, given a row (array of columns) for the
 * factoring over the factor base of length len, and the remainder when those
 * are divided away. */
static prtree *prtree_create_node(unsigned int *column, int len, mpz_t r,
mpz_t num) {
	prtree *prt;

	prt = malloc(sizeof(*prt) + len*sizeof(*column));
	prt->left = prt->right = NULL;
	prt->len = len;
	mpz_init_set(prt->r, r);
	mpz_init_set(prt->num, num);
	memcpy(prt->col, column, len*sizeof(*column));
	return prt;
}

/* Deallocate a single node without considering its children */
static void prtree_destroy_node(prtree *prt) {
	mpz_clear(prt->r);
	mpz_clear(prt->num);
	free(prt);
}

/* Deallocate a subtree recursively */
static void prtree_destroy(prtree *prt) {
	if(prt == NULL) return;
	if(prt->left != NULL) prtree_destroy(prt->left);
	if(prt->right != NULL) prtree_destroy(prt->right);
	prtree_destroy_node(prt);
}

/* Try to add the given partial relation to the tree with the given root.
 * If there is no partial relation in the tree with the same remainder, a new
 * node is created and NULL is returned.
 * Otherwise, the node containing the old partial relation is returned, so
 * that it can be combined with the new one. This is the responsibility of the
 * caller.
 */
static prtree *prtree_add_pr(prtree **root, unsigned int *col, int len,
mpz_t r, mpz_t num) {
	int rel;
	if(*root == NULL) {
		*root = prtree_create_node(col, len, r, num);
		return NULL;
	}
	rel = mpz_cmp(r, (*root)->r);
	if(rel == 0) {
		return *root;
	} else if(rel < 0) {
		return prtree_add_pr(&((*root)->left), col, len, r, num);
	} else {
		return prtree_add_pr(&((*root)->right), col, len, r, num);
	}
}

/* Combine a new partial relation with a prtree node. num should be the square
 * root mod n of the new relation. */
static void prtree_combine(struct sparse *matrix, prtree *prt,
unsigned int *newcol, int newlen, mpz_t num, mpz_t *p_inv, mpz_t n) {
	int len = 0, i0 = 0, i1 = 0, i;
	unsigned int col[newlen + prt->len];
	mpz_t r_inv;

	mpz_init(r_inv);
	mpz_invert(r_inv, prt->r, n);
	mpz_mul(num, num, r_inv);
	mpz_mod(num, num, n);
	mpz_mul(num, num, prt->num);
	mpz_mod(num, num, n);

	while(i0 < prt->len || i1 < newlen) {
		if(i0 < prt->len && i1 < newlen) {
			if(prt->col[i0] < newcol[i1]) {
				col[len++] = prt->col[i0++];
			} else if(prt->col[i0] > newcol[i1]) {
				col[len++] = newcol[i1++];
			} else {
				mpz_mul(num, num, p_inv[newcol[i1]]);
				mpz_mod(num, num, n);
				i0++;
				i1++;
			}
		} else if(i0 < prt->len) {
			while(i0 < prt->len) col[len++] = prt->col[i0++];
			break;
		} else {
			while(i1 < newlen) col[len++] = newcol[i1++];
			break;
		}
	}

	/* Add the new row.
	 * TODO: len = 0 corresponds to a congruence of squares. Although an
	 * unlikely event, we probably shouldn't just throw that away. */
	if(len > 0) {
		struct row *row = create_row();
		for(i=0; i<len; i++) append_element(row, col[i]);
		append_row(matrix, row);
	}

	mpz_clear(r_inv);
}
#endif /* USE_PR */

/* Main sieving function.
 * n		number to be factored. n = 1 mod 8 (use a multiplier if
 * 		necessary)
 * n_mult	said multiplier, may be 1
 * poly		"polynomial" struct, contains the state of the sieving
 * x		starting point of sieve, relative to poly->x0
 * x_max	ending point ...
 * dx		Sieing interval size, should be between 2^10 and 2^20 on most
 *		machines. Must divide x and x_max. Must be a multiple of 8.
 * num		Array of modular square roots corresponding to the smooth
 * 		numbers
 * matrix	Matrix representing the factoring of smooth numbers
 * max_rows	Max size of matrix (number of relations needed)
 * prt		Root of tree to store partial relations
 */
static void find_smooth(mpz_t n, unsigned long n_mult, polynomial *poly,
long x, long x_max, const int dx, mpz_t *num, struct sparse *matrix,
int max_rows, prtree **prt) {
	long *x_cand;
	unsigned long *n_sqrt_p;
	int smooth, i, j, power;
	mpz_t q, sqrt_n, tmp, a4;
	struct row *row;
	int_log log2_q[dx];
	int_log ilog;
	unsigned long logref;
	unsigned long abs_x;
	int log2_x;
	int treshold;
	uint64_t mask;

	unsigned long *p = poly->p;
	int_log *log2_p = poly->log_p;
	int plen = poly->plen;

	mpz_init(q);
	mpz_init(sqrt_n);
	mpz_init(tmp);
	mpz_init(a4);
	mpz_mul_2exp(a4, poly->a, 2);
	mpz_sqrt(sqrt_n, n);
	mpz_add_ui(sqrt_n, sqrt_n, 1);

#if USE_PR
	prtree *prnode;
	unsigned long pr_limit = 0x40*poly->p[poly->plen-1];
#endif

	/* Just to be on the safe side, let's allocate the largest array that
	 * could possibly be used. Normally only a few bytes are used. */
	x_cand = malloc((x_max-x) * sizeof(*x_cand));

	polynomial_roots(poly, x);
	n_sqrt_p = poly->root;

	/* Size in bits of largest remaining number that we bother to factor.
	 * Should be higher if partial relations are used. */
#if USE_PR
	treshold = log2_p[plen-1] + 11;
#else
	treshold = log2_p[plen-1] + 4;
#endif

repeat:;
	smooth = 0;
	/* Fill the log2_q array (of length dx) with integer approximations of
	 * log2(Q(x)) ~ 1 + log2(a) + log2(x0) + log2(x)
	 *
	 * Since we started using negative x, we have to make sure that if x is
	 * negative (sign_x < 0), then the logarithm will decrease. This code
	 * assumes that x does not cross 0 in a sieving interval.
	 */
	abs_x = (x < 0)? -x : x;
	log2_x = 0;
	while(abs_x) {
		abs_x /= 2;
		log2_x++;
	}
	ilog = mpz_sizeinbase(poly->a, 2) + mpz_sizeinbase(poly->x0, 2) +
		log2_x + 2 - treshold;
	logref = 3 << log2_x;
	if(x >= 0 && x < dx) {
		for(i=0;;) {
			if(logref >= dx) {
				memset(log2_q+i, ilog, dx-i);
				break;
			} else {
				memset(log2_q+i, ilog, logref-i);
				i = logref;
			}
			logref *= 2;
			ilog++;
		}
	} else if(x < 0 && x >= -dx) {
		for(i=0;;) {
			if(logref >= dx) {
				memset(log2_q+i, ilog, dx-i);
				break;
			} else {
				memset(log2_q+i, ilog, logref-i);
				i = logref;
			}
			logref /= 2;
			ilog--;
		}
	} else {
		memset(log2_q, ilog, dx);
	}
	
	/* Now, perform the actual sieving process. For all multiples of p[i]
	 * (starting from sqrt_p[i*2] and sqrt_p[i*2+1]) below qlen, subtract
	 * log2_p[i] from the relevant value in log2_q.
	 *
	 * The loop is split into two parts, one for small primes and one for
	 * large primes. The first combines the inner loops into one, while the
	 * second (for primes near or above dx) will never loop much.
	 */
	for(i=2; i<0x60 && i<plen; i++) {
		const long inc = p[i];
		const int_log log_p = log2_p[i];
		int j0 = poly->root[i*2], j1 = poly->root[i*2+1];
		int diff;
		if(j0 < j1) {
			diff = j1 - j0;
			while(j1 < dx) {
				log2_q[j1-diff] -= log_p;
				log2_q[j1] -= log_p;
				j1 += inc;
			}
			poly->root[i*2+1] = j1 - dx;
			if(j1 - diff < dx) {
				log2_q[j1-diff] -= log_p;
				j1 += inc;
			}
			poly->root[i*2] = j1 - diff - dx;
		} else {
			diff = j0 - j1;
			while(j0 < dx) {
				log2_q[j0-diff] -= log_p;
				log2_q[j0] -= log_p;
				j0 += inc;
			}
			poly->root[i*2] = j0 - dx;
			if(j0 - diff < dx) {
				log2_q[j0-diff] -= log_p;
				j0 += inc;
			}
			poly->root[i*2+1] = j0 - diff - dx;
		}
	}
	for(i=0x60; i<plen; i++) {
		const long inc = p[i];
		const int_log log_p = log2_p[i];
		for(j=poly->root[i*2]; j<dx; j+=inc) 
			log2_q[j] -= log_p;
		poly->root[i*2] = j - dx;
		for(j=poly->root[i*2+1]; j<dx; j+=inc)
			log2_q[j] -= log_p;
		poly->root[i*2+1] = j - dx;
	}

	/* Try to check for smooth values. A value is smooth if the remaining
	 * factor's 2-logarithm is small enough to be considered a rounding
	 * error and/or a partial relation.
	 */
	for(i=0; i<dx; i+=8) {
		mask = *(uint64_t*)(log2_q+i);
		if(!(mask & 0x8080808080808080ULL)) continue;
		for(j=i; j<i+8; j++) {
			if(log2_q[j] & 0x80) {
				x_cand[smooth++] = x + j;
			}
		}
	}

	/* For each of the returned candidates, try to factor it over the factor
	 * base. */
	for(i=0; i<smooth; i++) {
		row = create_row();

		if(x_cand[i] < 0) mpz_sub_ui(q, poly->x0, -x_cand[i]);
		else mpz_add_ui(q, poly->x0, x_cand[i]);
		mpz_mul_2exp(q, q, 1);
		mpz_mul(q, q, poly->a);
		mpz_add(q, q, poly->b);
		mpz_mul(tmp, q, poly->d2_inv);
		mpz_mod(tmp, tmp, n);
		mpz_set(num[max_rows - matrix->row_count - 1], tmp);
		mpz_pow_ui(q, q, 2);
		mpz_sub(q, q, n);
		mpz_fdiv_q(q, q, a4);

		/* Factor out -1 */
		if(mpz_sgn(q) < 0) {
			append_element(row, 0);
			mpz_abs(q, q);
		}
		/* Factor out powers of 2 */
		power = 0;
		while(mpz_tstbit(q, 0) == 0) {
			power++;
			mpz_fdiv_q_2exp(q, q, 1);
			if(power == 2) {
				power = 0;
				mpz_mul(tmp,
					num[max_rows - matrix->row_count - 1],
					poly->p_inv[1]);
				mpz_mod(tmp, tmp, n);
				mpz_set(num[max_rows - matrix->row_count - 1],
					tmp);
			}
		}
		if(power % 2 == 1) append_element(row, 1);

		j = 2;
		/* Factor out all other primes */
		power = 0;
		while(j < plen && mpz_cmp_ui(q, 1)) {
			if(mpz_divisible_ui_p(q, p[j])) {
				power++;
				mpz_divexact_ui(q, q, p[j]);
				if(power == 2) {
					power = 0;
					mpz_mul(tmp,
					num[max_rows-matrix->row_count-1],
						poly->p_inv[j]);
					mpz_mod(tmp, tmp, n);
					mpz_set(
					num[max_rows - matrix->row_count - 1],
						tmp);
				}
			} else {
				if(power % 2 == 1) append_element(row, j);
				j++;
				power = 0;
			}
		}
		if(power % 2 == 1) append_element(row, j);
		
		/* If we're left with 1, the number was completely factored. */
		if(!mpz_cmp_ui(q, 1)) {
			/* Add the row to the matrix. */
			append_row(matrix, row);
		} else {
#if USE_PR
			if(mpz_cmp_ui(q, pr_limit) < 0 &&
			(n_mult == 1 || !mpz_divisible_ui_p(q, n_mult))) {
				unsigned int *this_col, this_len;
				this_len = row_length(row);
				this_col = malloc(
					this_len * sizeof(*this_col));
				j = 0;
				struct element *e;
				for(e = row->first; e != NULL; e = e->next) {
					this_col[j++] = e->column;
				}
				mpz_set(tmp,
					num[max_rows - matrix->row_count - 1]);
				prnode = prtree_add_pr(prt,
					this_col, this_len, q, tmp);
				if(prnode == NULL) {
					/* It couldn't be factored, and there
					 * is no matching partial relation. */
				} else {
					/* Combine this partial relation with
					 * an old one */
					prtree_combine(matrix, prnode,
						this_col, this_len, tmp,
						poly->p_inv, n);
					mpz_set(
					num[max_rows - matrix->row_count],
						tmp);
				}
				free(this_col);
			}
#endif /* USE_PR */
			destroy_row(row);
		}
		if(matrix->row_count >= max_rows) break;
	}
	x += dx;
	if(x < x_max && matrix->row_count < max_rows) goto repeat;

	free(x_cand);
	mpz_clear(sqrt_n);
	mpz_clear(a4);
	mpz_clear(q);
	mpz_clear(tmp);
}

/* Perform gaussian elimination on the given rows, which are treated as bit
 * arrays with elements mod 2. Dividde out factor base primes from the num
 * array as they are eliminated (p_inv is a lookup table for the inverses of
 * the primes mod n). */
static int gauss(mpz_t *num, mpz_t *row, int rows, int cols, mpz_t n,
mpz_t *p_inv) {
	int i,j,best_row;
	mpz_t x, both;
	int pivot[rows];
	unsigned long bit, weight, best_weight;

	mpz_init(x);
	mpz_init2(both, cols);

	pivot[0] = cols-1;
	for(i=0; i<rows; i++) {
		/* Compute our initial guess for the pivot element column of
		 * row i */
		if(i > 0) pivot[i] = pivot[i-1] - 1;
		if(pivot[i] < 0) pivot[i] = -1;

		/* Find a row with the pivot element set. If the column is 0,
		 * increase the pivot column and try again. If several rows
		 * have the pivot element set, the one with the fewest bits
		 * set is chosen. */
		while(pivot[i] >= 0) {
			best_weight = ULONG_MAX;
			best_row = i;
			for(j=i+1; j<rows; j++) {
				if(mpz_tstbit(row[j], pivot[i])) {
					weight = mpz_popcount(row[j]);
					if(weight < best_weight) {
						best_weight = weight;
						best_row = j;
					}
				}
			}
			if(best_weight != ULONG_MAX) {
				mpz_swap(row[i], row[best_row]);
				mpz_swap(num[i], num[best_row]);
				break;
			}
			pivot[i]--;
		}

		/* Return in case we've left with only zero rows */
		if(pivot[i] < 0) {
			mpz_clear(x);
			mpz_clear(both);
			return i;
		}

		/* Eliminate the pivot column in other rows */
		for(j=i+1; j<rows; j++) {
			if(mpz_tstbit(row[j], pivot[i])) {
				mpz_and(both, row[i], row[j]);
				mpz_xor(row[j], row[j], row[i]);
				bit = 0;
				mpz_mul(x, num[j], num[i]);
				mpz_mod(x, x, n);
				while((bit=mpz_scan1(both, bit))!=ULONG_MAX) {
					mpz_mul(x, x, p_inv[bit]);
					mpz_mod(x, x, n);
					bit++;
				}
				mpz_set(num[j], x);
			}
		}
	}

	mpz_clear(x);
	mpz_clear(both);
	return i;
}

/* Count the number of columns in the specified matrix. */
static int count_columns(struct sparse *matrix) {
	int column_count = 0;
	struct element *e;
	for(e = matrix->first_row->first; e != NULL; e = e->next) { 
		if(column_count <= e->column) column_count = e->column + 1;
	}
	return column_count;
}

/* Perform structured gaussian elimination on the specified matrix. Also make 
 * according modifications to the numbers corresponding to each row and the
 * inverse of the primes corresponding to each column.
 */
static void structured_gauss(struct sparse *matrix, mpz_t *num, int num_len, 
mpz_t *p_inv, int p_inv_len) {
	unsigned int i, j, column_removed, column_count = 0;
	struct element *e;
	struct row *row;
	/* Records rows to be deleted. Index i contains a non-NULL pointer iff
	 * that pointer leads to row i which should be deleted. */
	struct row *rows_to_remove[num_len];
	/* Keeps track of which columns have been removed (0 if not, non-zero 
	 * otherwise). */
	unsigned int columns_removed[p_inv_len]; 
	
	/* Delete all columns that have a single non-zero coefficient and the
	 * rows in which those columns have non-zero coefficients.
	 */
	column_count = count_columns(matrix);
	unsigned int column_element_count[column_count];
	unsigned int column_subtraction[column_count];
	struct element *first_element_of_column[column_count];
	column_removed = 1;
	while(column_removed) {
		column_removed = 0;
		for(i = 0; i < num_len; i++) {
			rows_to_remove[i] = NULL;
		}
		for(i = 0; i < p_inv_len; i++) {
			columns_removed[i] = 0;
		}
		
		for(i = 0; i < column_count; i++) {
			column_element_count[i] = 0;
			column_subtraction[i] = 0;
			first_element_of_column[i] = NULL;
		}

		/* Count the non-zero elements in each column. */
		for(e = matrix->first_row->first; e != NULL; e = e->next) {
			if(first_element_of_column[e->column] == NULL) {
				first_element_of_column[e->column] = e;
			}
			column_element_count[e->column]++;
		}

		/* A column is deleted by decreasing the column value of all
		 * elements in column greater than the deleted column. The
		 * value to decrease a given  column with is stored in
		 * column_subtraction. */
		for(i = 0; i < column_count; i++) {
			if(column_element_count[i] == 1) {
				column_removed = 1;
				columns_removed[i] = 1;
				/* Increse subtraction for greater columns. */
				for(j = i + 1; j < column_count; j++) {
					column_subtraction[j]++;
				}
				
				/* Mark the corresponding row for deletion. */
				row = first_element_of_column[i]->row;
				rows_to_remove[row->index] = row;
			}
		}

		/* Apply the subtraction. */
		for(e = matrix->first_row->first; e != NULL; e = e->next) {
			e->column -= column_subtraction[e->column];
		}
		
		/* Remove the row numbers that are no longer used. */
		j = 0; /* The next location to assign to.  */
		for(i = 0; i < num_len; i++) {
			if(rows_to_remove[i] == NULL) {
				if(i != j) mpz_swap(num[j], num[i]);
				j++;
			} else {
				/* Modify the row indices. */
				for(row = rows_to_remove[i]->next; row != NULL; 
						row = row->next) {
					row->index--;
				}
				
				/* Remove the row. */
				row = rows_to_remove[i];
				if(matrix->first_row == row) {
					matrix->first_row = row->next;
				}
				if(matrix->last_row == row) {
					matrix->last_row = row->prev;
				}
				matrix->row_count--;
				destroy_row(row);
			}
		}
		
		/* Remove the prime inverses that are no longer used. */
		j = 0; /* The next location to assign to. */
		for(i = 0; i < p_inv_len; i++) {
			if(!columns_removed[i]) {
				if(i != j) mpz_swap(p_inv[j], p_inv[i]);
				j++;
			} else {
				column_count--;
			}
		}
	}
}

/* Reverses the rows in the specified matrix. */
static void reverse_rows(struct sparse *matrix) {
	/* Disconnect each row from the original matrix and reattach them to
	 * a new matrix in reverse order, then reattach the original matrix.
	 */
	struct sparse *reversed = create_sparse();
	struct row *r, *prev;
	for(r = matrix->last_row; r != NULL; r = prev) {
		prev = r->prev;
		r->prev = NULL;
		r->next = NULL;
		r->last->next = NULL;
		append_row(reversed, r);
	}
	
	matrix->first_row = reversed->first_row;
	matrix->last_row = reversed->last_row;
	
	reversed->first_row = NULL;
	reversed->last_row = NULL;
	free(reversed);
}

/* Attempt to find a number r that divides n. Return 0 on success (i.e. a
 * non-trivial r was found), 1 otherwise. */
int gmp_qs_factor(mpz_t r, mpz_t n0) {
	size_t n_size;
	double ln_n;
	int B;
	mpz_t n, n_1, x;
	unsigned int log2_dx = SIEVE_LOG_STEP, dx;
	long M = SIEVE_LEN;
	int i, first_row, column_count, retval = 1;
	unsigned long n_mult;

	/* We need to ensure that n = 1 (mod 8), which could mean multiplying
	 * it with a small constant. We will set n_mult to this constant. */
	mpz_init(n);
	mpz_fdiv_r_2exp(n, n0, 3);
	if(!mpz_cmp_ui(n, 1)) {
		/* If n = 1 (mod 8), do nothing */
		n_mult = 1;
		mpz_set(n, n0);
	} else {
		/* Otherwise, let  n' = (n mod 8) * n */
		n_mult = mpz_get_ui(n);
		mpz_mul(n, n, n0);
	}

	dx = 1 << log2_dx;

	n_size = mpz_sizeinbase(n, 2);
	ln_n = M_LN2 * (double)n_size;

	/* This seems to be close to optimal for all numbers tested */
	B = (int) (0.63*pow(exp(sqrt(ln_n * log(ln_n))), 0.35355339059));

	int rows = B + 16;
	mpz_t num[rows];
	for(i=0; i<rows; i++) mpz_init2(num[i], n_size);

	struct sparse *matrix = create_sparse();

	polynomial *poly;
	mpz_t d;
	poly = polynomial_create(n, B);
	mpz_init(d);
	mpz_set_ui(d, B);
	mpz_pow_ui(d, d, 6);
	mpz_mul_2exp(d, d, 1);
	mpz_fdiv_q(d, n, d);
	mpz_sqrt(d, d);
	mpz_sqrt(d, d);

#if USE_PR
	prtree *prt = NULL;
	while(matrix->row_count < rows) {
		polynomial_update(poly, d, n);
		find_smooth(
			n, n_mult, poly, -M, M, dx, num, matrix, rows, &prt);
	}
	prtree_destroy(prt);
#else
	while(matrix->row_count < rows) {
		polynomial_update(poly, d, n);
		find_smooth(
			n, n_mult, poly, -M, M, dx, num, matrix, rows, NULL);
	}
#endif

	/* Reverse matrix to stay be in the same order as num and p_inv. */
	reverse_rows(matrix);
	structured_gauss(matrix, num, rows, poly->p_inv, B);
	
	/* Convert the sparse matrix to a full matrix representation. */
	mpz_t row[matrix->row_count];
	column_count = count_columns(matrix);
	struct element *element;
	struct row *rowp;
	for(i = 0; i < matrix->row_count; i++) {
		mpz_init2(row[i], column_count);
	}
	i = -1;
	rowp = matrix->first_row;
	for(element = matrix->first_row->first; element != NULL; 
			element = element->next) {
		if(rowp != NULL && rowp->first == element) {
			i++;
			rowp = rowp->next;
		}
		mpz_setbit(row[i], element->column);
	}

	int row_count = matrix->row_count;
	destroy_sparse(matrix);

	first_row = gauss(num, row, row_count, column_count, n,
		poly->p_inv);

	mpz_init(n_1);
	mpz_init(x);
	mpz_sub_ui(n_1, n0, 1);
	for(i=first_row; i<row_count; i++) {
		if(!mpz_cmp_ui(num[i], 1)) continue;
		if(!mpz_cmp(num[i], n_1)) continue;
		mpz_sub_ui(x, num[i], 1);
		mpz_gcd(x, x, n0);
		if(!mpz_cmp_ui(x, 1)) continue;
		if(!mpz_cmp(x, n0)) continue;
		retval = 0;
		mpz_set(r, x);
		break;
	}

	polynomial_destroy(poly);
	mpz_clear(n);
	mpz_clear(x);
	mpz_clear(d);
	mpz_clear(n_1);
	for(i=0; i<rows; i++) mpz_clear(num[i]);
	for(i=0; i<row_count; i++) mpz_clear(row[i]);
	
	return retval;
}

