/* gmp_rho.c
 *
 * Implementations of Pollard's rho algorithm.
 * The original version using both 64-bit arithmetic (for numbers < 2^32)
 * and using the GNU MP library, as well as Brent's version for GNU MP.
 */

#include "gmp_rho.h"

#define min(x, y) ((x < y)? x : y)

static uint64_t gcd64(uint64_t x, uint64_t y) {
	uint64_t t;

	if(x > y) {
		while(y) {
			t = y;
			y = x % y;
			x = t;
		}
		return x;
	}
	while(x) {
		t = x;
		x = y % x;
		y = t;
	}
	return y;
}

static int factor64_rho_c(
uint64_t *d, uint64_t n, uint64_t c, unsigned long *max) {
	uint64_t x, y, diff_xy, g;

	x = y = 4 + c;
	for(;;) {
		if(max != NULL && (*max)-- == 0) return -1;
		y = (y*y + c) % n;
		diff_xy = (x > y)? x - y : y - x;
		g = gcd64(diff_xy, n);
		if(g == n) return 0;
		if(g > 1) { *d = g; return 1; }
		x = (x*x + c) % n;
		y = (y*y + c) % n;
	}
}

/* Factor the number n into a (not necessarily prime) factor p. This will run
 * until one such divisor of n is found, which in the case  of a prime number
 * means forever.
 * At most max iterations are performed. max = 0 disables this limit.
 * Return 0 on success, or 1 on failure (out of time).
 */
int factor64_rho(uint64_t *p, uint64_t n, unsigned long max) {
    uint64_t c = 1;
	int retval;

	for(;;) {
		if(max) retval = factor64_rho_c(p, n, c, &max);
		else retval = factor64_rho_c(p, n, c, NULL);
		if(retval < 0) return 1;
		if(retval > 0) return 0;
		c++;
	}
}

static void mpz_square_add_mod(
mpz_t rop, mpz_t op1, unsigned int op2, mpz_t m) {
    mpz_powm_ui(rop, op1, 2, m);        /* rop = op1^2 (mod m)              */
    mpz_add_ui(rop, rop, op2);          /* rop = rop + op2                  */
    if(mpz_cmp(rop, m) >= 0)
        mpz_sub(rop, rop, m);           /* Last addition mod m              */
}

/* Try to break up n into (not necessarily prime) factors, using Pollard's rho
 * algorithm with the pseudorandom function f(x) = x^2 + c (mod n). 0 will be
 * returned if no factor could be found, in which case you want to change c
 * and try again. 1 is returned if d contains a number which divides n.
 */
static int gmp_factor_rho_c(
mpz_t d, mpz_t n, unsigned int c, unsigned long *max) {
    mpz_t x, y, diff_xy;

    mpz_init_set_ui(x, 4+c);            /* x0 = 2, x1 = 2^2+c               */
    mpz_init_set_ui(y, 4+c);            /* y0 = 2, y1 = 2^2+c               */
    mpz_init(diff_xy);
    for(;;) {                           /* While d = 1:                     */
		if(max != NULL) {
			if((*max)-- == 0) {
		        mpz_clear(x);
    		    mpz_clear(y);
           		mpz_clear(diff_xy);
				return -1;
			}
		}
        mpz_square_add_mod(y, y, c, n); /*  y = y^2 + c (mod n)             */
        mpz_sub(diff_xy, x, y);
        mpz_abs(diff_xy, diff_xy);      /*  diff_xy = abs(x-y)              */
        mpz_gcd(d, diff_xy, n);         /*  d = gcd(diff_xy, n)             */
        if(!mpz_cmp(d, n)) {
            mpz_clear(x);
            mpz_clear(y);
            mpz_clear(diff_xy);
            return 0;                   /*  d = n? Then return failure      */
        }
        if(mpz_cmp_ui(d, 1) > 0) {
            mpz_clear(x);
            mpz_clear(y);
            mpz_clear(diff_xy);
            return 1;                   /*  d = 1? Then return success      */
        }
        mpz_square_add_mod(x, x, c, n); /*  x = x^2 + c (mod n)             */
        mpz_square_add_mod(y, y, c, n); /*  y = y^2 + c (mod n)             */
    }
}

/* Factor the number n into a (not necessarily prime) factor p. This will run
 * until one such divisor of n is found, which in the case  of a prime number
 * means forever.
 * At most max iterations are performed. max = 0 disables this limit.
 * Return 0 on success, or 1 on failure (out of time).
 */
int gmp_factor_rho(mpz_t p, mpz_t n, unsigned long max) {
    unsigned int c = 1;
	int retval;

	for(;;) {
		if(max) retval = gmp_factor_rho_c(p, n, c, &max);
		else retval = gmp_factor_rho_c(p, n, c, NULL);
		if(retval < 0) return 1;
		if(retval > 0) return 0;
		c++;
	}
}

static int gmp_factor_brent_c(
mpz_t g, mpz_t n, unsigned int c, unsigned long *max) {
    mpz_t x, y, ys, diff_xy, q;
    unsigned int i, j, r, m, k;

    mpz_init(x);
    mpz_init(ys);
    mpz_init(diff_xy);
    m = 1;                              /* m = 11                           */
    r = 1;                              /* r = 1                            */
    mpz_init_set_ui(q, 1);              /* q = 1                            */
    mpz_init_set_ui(y, 4+c);            /* y = 2^2 + c                      */
    do {                                /* Do:                              */
        mpz_set(x, y);                  /*    x =                           */
        k = 0;                          /*    k = 0                         */
        for(i=0; i<r; i++)              /*    Do r times:                   */
            mpz_square_add_mod(y, y, c, n); /*    y = y^2 + c (mod n)       */
        do {
			if(max != NULL) if((*max)-- == 0) return -1;
            mpz_set(ys, y);             /*    ys = y                        */
            j = min(m, r-k);
            for(i=0; i<j; i++) {        /*    Do min(m, r-k) times:         */
                mpz_square_add_mod(y, y, c, n); /* y = y^2 + c (mod n)      */
                mpz_sub(diff_xy, x, y);
                mpz_abs(diff_xy, diff_xy);
                mpz_mul(q, q, diff_xy); /*         q = q * abs(x-y) (mod n) */
                mpz_mod(q, q, n);
            }
            mpz_gcd(g, q, n);           /*         g = gcd(q, n)            */
            k = k + m;                  /*         k = k + m                */
                                        /*    Until k >= r, or g > 1        */
        } while(k < r && mpz_cmp_ui(g, 1) <= 0);
        r = r * 2;                      /*    r = r * 2                     */
    } while(mpz_cmp_ui(g, 1) <= 0);     /* Until g > 1                      */
    if(!mpz_cmp(g, n)) {                /* If g = n:                        */
        do {                            /*    Do:                           */
            mpz_square_add_mod(ys, ys, c, n); /* ys = ys^2 + c (mod n)      */
            mpz_sub(y, x, ys);          /*       y = x - ys                 */
            mpz_gcd(g, y, n);           /*       g = gcd(x - ys, n)         */
        } while(mpz_cmp_ui(g, 1) <= 0); /*    until g > 1                   */
    }
    mpz_clear(x);
    mpz_clear(y);
    mpz_clear(ys);
    mpz_clear(diff_xy);
    mpz_clear(q);
    if(!mpz_cmp(g, n)) return 0;
    return 1;
}

int gmp_factor_brent(mpz_t p, mpz_t n, unsigned long max) {
    unsigned int c = 1;
	int retval;

	for(;;) {
		if(max) retval = gmp_factor_brent_c(p, n, c, &max);
		else retval = gmp_factor_brent_c(p, n, c, NULL);
		if(retval < 0) return 1;
		if(retval > 0) return 0;
		c++;
	}
}

