/*
    Copyright (C) 2007  Robert Östling and Andreas Launila

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "gmp_rho.h"
#include "gmp_qs.h"
#include "gmp_p1.h"
#include "gmp_pm1.h"
#include "primes.h"

static enum {
	MODE_DEFAULT,	/* Default algorithms (rho and QS) */
	MODE_PM1,	/* Pollard's p-1 algorithm */
	MODE_P1,	/* Williams' p+1 algorithm */
	MODE_RHO	/* Pollard's rho algorithm */
} mode = MODE_DEFAULT;

/* Attempt to find a factor of n.
 * Return k < 0 if unable to factor.
 * Return 0 if n is (probably) prime.
 * Return 1 if one factor if n is found. The factor is written to r.
 * Return k > 1 if there are k and m such that n = m^k. m is written to r.
 */
static int find_factor(mpz_t r, mpz_t n) {
	mpz_t p;
	size_t n_size;
	int retval;

	/* If n is prime, set r to the n itself */
	if(mpz_probab_prime_p(n, 10)) {
		mpz_set(r, n);
		return 0;
	}

	/* If n is a perfect square, set r to its square root and return 2 */
	if(mpz_perfect_square_p(n)) {
		mpz_sqrt(r, n);
		return 2;
	}

	/* If n is a perfect (odd) power, attempt to find the base */
	if(mpz_perfect_power_p(n)) {
		mpz_init_set_ui(p, 3);
		for(;;) {
			if(mpz_root(r, n, mpz_get_ui(p))) break;
			mpz_nextprime(p, p);
		}
		retval = mpz_get_ui(p);
		mpz_clear(p);
		return retval;
	}

	switch(mode) {
	case MODE_PM1:
		if(gmp_factor_pm1(r, n, 4, 0x1000)) return -1;
		return 1;
	case MODE_P1:
		if(gmp_factor_p1(r, n, 0x1000, 4)) return -1;
		return 1;
	case MODE_RHO:
		gmp_factor_rho(r, n, 0);
		return 1;
	default:
		break;
	}

	n_size = mpz_sizeinbase(n, 2);

	/* For really small numbers, use the fixed-precision version of
	 * Pollard's rho algorithm. */
	if(n_size <= 32) {
		uint64_t r64;
		factor64_rho(&r64, mpz_get_ui(n), 0);
		mpz_set_ui(r, r64);
		return 1;
	}

	/* For small numbers, use the general rho algorithm */
	if(n_size < 45) {
		gmp_factor_rho(r, n, 0);
		return 1;
	}

	/* Try QS. If this fails, give up. */
	if(gmp_qs_factor(r, n)) {
		return -1;
	}

	return 1;
}

/* Print every prime factor of n reps times. Return 0 on success. */
static int print_large_factors(mpz_t n, int reps) {
	mpz_t r;
	int i, power;

	mpz_init(r);
	power = find_factor(r, n);
	if(power < 0) {
		mpz_clear(r);
		return 1;
	}
	if(power == 0) {
		for(i=0; i<reps; i++) {
			gmp_printf("%Zd\n", r);
		}
	} else {
		for(i=0; i<power; i++) mpz_divexact(n, n, r);
		print_large_factors(r, reps*power);
		if(mpz_cmp_ui(n, 1) > 0) print_large_factors(n, reps);
	}
	mpz_clear(r);
	return 0;
}

/* Find all prime factors of n and print them to stdout. */
static int print_factors(mpz_t n) {
	int i;

	/* Factor out powers of two */
	while(mpz_even_p(n)) {
		puts("2");
		mpz_fdiv_q_2exp(n, n, 1);
		if(!mpz_cmp_ui(n, 1)) return 0;
	}

	/* Try dividing with some small primes */
	for(i=0; i<PRIMES; i++) {
		while(mpz_divisible_ui_p(n, prime_table[i])) {
			printf("%lu\n", prime_table[i]);
			mpz_divexact_ui(n, n, prime_table[i]);
			if(!mpz_cmp_ui(n, 1)) return 0;
		}
	}

	return print_large_factors(n, 1);
}

int main(int argc, char *argv[]) {
	int c, i;
	mpz_t n;

	for(;;) {
		if((c = getopt(argc, argv, "mpr")) == -1) break;
		switch(c) {
		case 'm':
			mode = MODE_PM1; break;
		case 'p':
			mode = MODE_P1; break;
		case 'r':
			mode = MODE_RHO; break;
		default:
			fprintf(stderr, "%s: unknown option\n", argv[0]);
			exit(1);
		}
	}

	mpz_init(n);

	if(argc > optind) {
		for(i=optind; i<argc; i++) {
			mpz_set_str(n, argv[i], 0);
			if(print_factors(n)) {
				printf("fail\n");
				exit(1);
			}
			putchar('\n');
			fflush(stdout);
		}
		mpz_clear(n);
		return 0;
	}

	for(;;) {
		if(gmp_scanf("%Zd", n) != 1) break;
		if(print_factors(n)) {
			printf("fail\n");
			exit(1);
		}
		putchar('\n');
		fflush(stdout);
	}

	mpz_clear(n);
	return 0;
}

