#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <gmp.h>

/* Generate a random n-bit prime, using the random state s.
 * The type t can be:
 *
 *   0: No requirements (b is ignored)
 *   1: p - 1 is b-smooth
 *   2: p + 1 is b-smooth
 */
void find_prime(gmp_randstate_t s, mpz_t p, int n, int t, mpz_t b) {
	mpz_t q;
	size_t bits, b_bits;

	if(t == 0) {
		mpz_urandomb(p, s, n);
		mpz_setbit(p, n-1);
		mpz_nextprime(p, p);
		return;
	}
	if(t >= 1 && t <= 2) {
		mpz_init(q);
		b_bits = mpz_sizeinbase(b, 2);
		for(;;) {
			mpz_set_ui(p, 1);
			while((bits = mpz_sizeinbase(p, 2)) < n) {
				if(n - bits < b_bits) {
					mpz_urandomb(q, s, n - bits);
					mpz_setbit(q, n - bits - 1);
					mpz_nextprime(q, q);
				} else {
					do {
						mpz_urandomm(q, s, b);
						mpz_nextprime(q, q);
					} while(mpz_cmp(q, b) > 0);
				}
				mpz_mul(p, p, q);
			}
			if(t == 2) mpz_sub_ui(p, p, 1);
			else mpz_add_ui(p, p, 1);
			if(mpz_probab_prime_p(p, 10)) break;
		}
		mpz_clear(q);
		return;
	}
	fprintf(stderr, "find_prime: t = %d (invalid value)\n", t);
	abort();
}

void help(FILE *file) {
	fprintf(file,
	"sample [-h] [-m b | -p b] [-s] size1 [size2 [size3 ...]]\n\n"
	"    -h       Show this message\n"
	"    -m b     The first prime - 1 must be b-smooth\n"
	"    -p b     The first prime + 1 must be b-smooth\n"
	"    -s       Show factors (not just product)\n");
}

int main(int argc, char **argv) {
	mpz_t n, p, b;
	gmp_randstate_t state;
	int c, i, t, show;
	long bits;
	char *endptr;
	struct timeval tv;

	show = 0;
	t = 0;
	mpz_init(b);

	for(;;) {
		if((c = getopt(argc, argv, "hm:p:s")) == -1) break;
		switch(c) {
		case 'h':
			help(stdout);
			exit(0);
		case 'm':
			t = 1;
			mpz_set_str(b, optarg, 0);
			break;
		case 'p':
			t = 2;
			mpz_set_str(b, optarg, 0);
			break;
		case 's':
			show = 1;
			break;
		default:
			fprintf(stderr,
				"%s: unknown option (use -h for help)\n",
				argv[0]);
			exit(1);
		}
	}

	if(optind >= argc) {
		fprintf(stderr,
			"%s: no prime sizes given (use -h for help)\n",
			argv[0]);
		exit(1);
	}

	gmp_randinit_default(state);
	gettimeofday(&tv, NULL);
	gmp_randseed_ui(state, tv.tv_usec);

	mpz_init_set_ui(n, 1);
	mpz_init(p);

	for(i=optind; i<argc; i++) {
		bits = strtol(argv[i], &endptr, 0);
		if(*endptr) {
			fprintf(stderr, "%s: not a number: %s\n",
				argv[0], argv[i]);
			exit(1);
		}
		if(t >= 1 && t <= 2 && i == optind) {
			find_prime(state, p, atoi(argv[i]), t, b);
		} else {
			find_prime(state, p, atoi(argv[i]), 0, b);
		}
		mpz_mul(n, n, p);
		if(show) {
			if(i == optind) gmp_printf("%Zd", p);
			else gmp_printf(" * %Zd", p);
		}
	}

	if(show) printf(" = ");
	gmp_printf("%Zd\n", n);

	mpz_clear(b);
	mpz_clear(n);
	mpz_clear(p);
	gmp_randclear(state);

	return 0;
}

